#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "bollettino-neve.pb-c.h"
#define DIM 256

int main(int argc, char** argv){
	int err;
	struct addrinfo hints;
	struct addrinfo *res,*ptr;
	char *host_remoto;
	char *servizio_remoto;
	int sd, nread;
	char buf[DIM], regione[DIM], c, dim[1];
	void *req_buf;
	int i=1, len;
	
	Richiesta req = RICHIESTA__INIT;

	/* Controllo argomenti */
	if (argc < 3){
		printf("Uso: bollettino_neve <server> <porta>\n");
		exit(1);
	}

	/* Lettura dati dall'utente */
	printf("Inserisci il nome della regione cui sei interessato:\n");
	scanf("%s", regione);
	req.nome_regione = regione;
	
	/* Lettura dati dall'utente */
	printf("Inserisci il numero di impianti da visualizzare:\n");
	scanf("%d", &(req.numero_impianti)); //ho il terminatore
	
	/* Costruzione dell'indirizzo */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	/* Risoluzione dell'host */
	host_remoto = argv[1];
	servizio_remoto = argv[2];

	if ((err = getaddrinfo(host_remoto, servizio_remoto, &hints, &res)) != 0) {
		fprintf(stderr, "Errore risoluzione nome: %s\n", gai_strerror(err));
		exit(2);
	}

	/* Connessione con fallback */
	for(ptr = res; ptr != NULL; ptr = ptr->ai_next){
		/*se socket fallisce salto direttamente alla prossima iterazione*/
		if ((sd = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol)) < 0){
			fprintf(stderr,"creazione socket fallita\n");
			continue;
		}
		/*se connect funziona esco dal ciclo*/
		if (connect(sd, ptr->ai_addr, ptr->ai_addrlen) == 0){
			printf("connect riuscita al tentativo %d\n",i);
			break;
		}
		i++;
		close(sd);
	}

	/* Verifica sul risultato restituito da getaddrinfo */
	if (ptr == NULL) {
		fprintf(stderr, "Errore risoluzione nome: nessun indirizzo corrispondente trovato\n");
		exit(3);
	}

	// Liberiamo la memoria allocata da getaddrinfo()
	freeaddrinfo(res);

	/* Scambio dati col server */
	
	// Serializzazione Protobuf
	len = richiesta__get_packed_size(&req);
	req_buf = malloc(len);
	richiesta__pack(&req,req_buf);
	
	// Invio la dimensione del messaggio Protobuf
	dim[0] = len;
	if (write(sd, dim, 1) < 0) { 
			perror("write dim[0]");
			exit(EXIT_FAILURE);
	}
	
	// Mando il messaggio Protobuf
	if (write(sd, req_buf, len) < 0) {
		perror("write struct");
		exit(5);
	}
	
	// Importante!
	free(req_buf);
	
	/* Ricezione risultato */

	/* È consigliabile effettuare il flushing del buffer di printf
	 * prima di iniziare a scrivere sullo standard output con write */
	fflush(stdout);

	memset(buf, 0, sizeof(buf));
	while ((nread = read(sd, buf, DIM)) > 0) {
		if (write(1, buf, nread) < 0) {
			 perror("write su stdout");
			 exit(9);
		}
	}

	/* Controllo errori di lettura */
	if (nread < 0) {
		perror("read del risultato");
		exit(10);
	}

	return 0;
}
