#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#include "lista-articoli.pb-c.h"
#define N 256

int ricevi (int sd, char* buf, int n){
   int i, j;
   i = recv (sd, buf, n, 0);
   if ( i != n && i != 0) {
      if (i == -1) { 
         fprintf(stderr,"errore in lettura\n");
          exit(1); }
      while (i < n) { 
         j = recv (sd, &buf[i], n-i, 0);
         if (j == -1) { 
            fprintf(stderr,"errore in lettura\n"); 
            exit(1); }
         i += j; 
         if (j == 0) break; 
      }
   } /* si assume che tutti i byte arrivino… se si verifica il fine file si esce */
   return i;
}

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	while (waitpid(-1, &status, WNOHANG) > 0);
}

int autorizza(char *id, int password)
{
	printf("autorizza: %s %d\n", id, password);
	return 1;
}

int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, pid, on, piped[2], ok, nread;
	
	Revisore *revisore;
	char buf[N], dim[1];
	char username_str[N], email[N];
	
	// Creando la stringa aggiungo \n per consentire al client Java la readLine
	char *access_denied = "accesso non autorizzato\n";

	/* Controllo argomenti */
	if(argc < 2){
		printf("Uso: ./server <porta> \n");
		exit(1);
	}

	signal(SIGCHLD, handler);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, argv[1], &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}

	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}

	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}

	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}

	freeaddrinfo(res);

	if(listen(sd, SOMAXCONN)<0){
		perror("listen");
		exit(6);
	}

	/* Attendo i client... */
	for(;;){
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}

		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		} else if(pid == 0){

			/* Chiudo la socket passiva */
			close(sd);

			/* figlio */
			
			/* Leggo un singolo byte con la dimensione del messaggio */
			if ((nread = read(ns, dim, 1)) < 0) {
				perror("read dim");
				exit(EXIT_FAILURE);
			}

			/* Leggo esattamente buff[0] byte. */
			if( ricevi(ns, buf, dim[0]) < 0 ){
				perror("read protobuf");
				exit(EXIT_FAILURE);
			}
	
			// Deserializzazione messaggio Protobuf
			revisore = revisore__unpack(NULL, dim[0], buf);
			if( revisore == NULL ){
				perror("deserializzazione protobuf");
				exit(9);
			}
			
			// verifica credenziali
			if( ! revisore->has_username ){
				ok = autorizza(revisore->email, revisore->password);
			}
			else{
				sprintf(username_str, "%d", revisore->username);
				ok = autorizza(username_str, revisore->password);
			}
			if ( ok != 1 )
			{
				if(write(ns, access_denied, strlen(access_denied))<0){
					perror("write access denied");
					exit(11);
				}
				close(ns);
				exit(13);
			}
			
			memcpy(email, revisore->email, strlen(revisore->email));
			
			// Ora posso rilasciare la memoria allocata in fase di deserializzazione
			revisore__free_unpacked(revisore, NULL);
			
			pipe(piped);
			if(fork()==0)
			{
				//nipote
				close(piped[0]);
				close(1);
				dup(piped[1]);
				close(piped[1]);
				close(ns);

				execlp("grep", "grep", email, "articoli.txt", (char *)0);
				perror("exec grep");
				exit(14);
			}
			//ancora figlio
			close(piped[1]);
			close(0);
			dup(piped[0]);
			close(piped[0]);

			/* Ridireziono stdout */
			close(1);
			dup(ns);
			close(ns);

			execlp("sort", "sort", "-r", "-n", (char *)0);
			perror("exec sort");
			exit(15);

		}
		// padre
		close(ns);
	}
	return 0;
}
