#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#define N 256

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	while (waitpid(-1, &status, WNOHANG) > 0);
}

int autorizza(char *id, int password)
{
	printf("autorizza: %s %d\n", id, password);
	return 1;
}

int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, pid, on, piped[2], ok;
	
	char username[N], email[N], password[N];
	
	// creando le stringhe aggiungo \n per consentire al client Java la readLine
	char *ack = "ack\n";
	char *access_denied = "accesso non autorizzato\n";

	/* Controllo argomenti */
	if(argc < 2){
		printf("Uso: ./server <porta> \n");
		exit(1);
	}

	signal(SIGCHLD, handler);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, argv[1], &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}

	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}

	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}

	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}

	freeaddrinfo(res);

	if(listen(sd, SOMAXCONN)<0){
		perror("listen");
		exit(6);
	}

	/* Attendo i client... */
	for(;;){
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}

		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		} else if(pid == 0){

			/* Chiudo la socket passiva */
			close(sd);

			/* figlio */
			
			/* Leggo lo username */
			memset(username, 0, sizeof(username));
			if(read(ns, username, N-1)<0){
				perror("read username");
				exit(9);
			}
			if(write(ns, ack, strlen(ack))<0){
				perror("write ack 1");
				exit(10);
			}
			
			/* Leggo l'email */
			memset(email, 0, sizeof(email));
			if(read(ns, email, N-1)<0){
				perror("read email");
				exit(9);
			}
			if(write(ns, ack, strlen(ack))<0){
				perror("write ack 2");
				exit(10);
			}
			
			memset(password, 0, sizeof(password));
			if(read(ns, password, N-1)<0){
				perror("read password");
				exit(11);
			}
			
			// verifica credenziali
			if( strcmp(username, "-1") == 0){
				ok = autorizza(email, atoi(password));
			}
			else{
				ok = autorizza(username, atoi(password));
			}
			if ( ok != 1 )
			{
				if(write(ns, access_denied, strlen(access_denied))<0){
					perror("write access denied");
					exit(11);
				}
				close(ns);
				exit(13);
			}

			pipe(piped);
			if(fork()==0)
			{
				//nipote
				close(piped[0]);
				close(1);
				dup(piped[1]);
				close(piped[1]);
				close(ns);

				execlp("grep", "grep", email, "articoli.txt", (char *)0);
				perror("exec grep");
				exit(14);
			}
			//ancora figlio
			close(piped[1]);
			close(0);
			dup(piped[0]);
			close(piped[0]);

			/* Ridireziono stdout */
			close(1);
			dup(ns);
			close(ns);

			execlp("sort", "sort", "-r", "-n", (char *)0);
			perror("exec sort");
			exit(15);

		}
		// padre
		close(ns);
	}
	return 0;
}
